# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2021 Osyris contributors (https://github.com/nvaytet/osyris)

# flake8: noqa

from .config import config, units
from .plot import histogram, plane, scatter
from .core import Array, Datagroup, Dataset, Plot
